/*
  Tutanchamun, an arcade game
  Copyright (C) 1998  David Kastrup
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
#include "tutan.h"
#include "tutsound.h"
#include "tutdisp.h"
#include "assert.h"

mcb mcbarr[13], *freemcb;
int nmons;

#define B_MOVED 1
#define B_KILLED 2
#define B_BORN 4
#define B_CLOUDY 8

void grabit(mcb *mp)
{
  undraw(mp->shape[(mp->flag & (1<<PHASE)) != 0], mp->xcoor, mp->ycoor);
}

void throwi(mcb *mp)
{
  if (mp->flag & 2)
    mp->shape = mp->mph->shaplf;
  else if (mp->flag & 1)
    mp->shape = mp->mph->shapri;
  draw(mp->shape[(mp->flag & (1<<PHASE)) != 0], mp->xcoor, mp->ycoor);
}

unsigned legal(mcb *mp, ubyte staga[][12])
{
  if (mp->xcoor & 3)
    return 3;
  if (mp->ycoor & 3)
    return 0xc;
  return (staga[mp->xcoor/4+1][mp->ycoor/4] & 0x1f ? 0 : 1)
    | (mp->xcoor == 0 || (staga[mp->xcoor/4-1][mp->ycoor/4] & 0x1f) ? 0 : 2)
    | (staga[mp->xcoor/4][mp->ycoor/4-1] & 0x1f ? 0 : 4)
    | (staga[mp->xcoor/4][mp->ycoor/4+1] & 0x1f ? 0 : 8);
}

unsigned newcoo(coor *coo, mcb *mp)
{
  coo->x = mp->xcoor;
  coo->y = mp->ycoor;
  if (--mp->downcn == 0) {
    mp->downcn = mp->mph->inidow;
    if (mp->flag & (1<<DEATH))
      return B_KILLED|B_MOVED;
    if (mp->flag & (1<<BIRTH)) {
      mp->flag &= ~(1<<BIRTH);
      return B_BORN|B_MOVED;
    }
    switch (mp->flag & 0xf) {
    case 1:
      coo->x += mp->mph->distan;
      break;
    case 2:
      coo->x -= mp->mph->distan;
      break;
    case 4:
      coo->y -= mp->mph->distan;
      break;
    case 8:
      coo->y += mp->mph->distan;
      break;
    case 0:
      return 0;
    default:
      assert(0);
    }
    return B_MOVED;
  }
  if (mp->flag & ((1<<BIRTH)|(1<<DEATH))) {
    if(mp->downcn & 1)
      return B_CLOUDY;
    return B_MOVED|B_CLOUDY;
  }
  return 0;
}

void gbonus(int quality, int x, int y, void *final)
{
  static scotab[6] = {5,10,15,20,30,40};
  draw(d_psh[quality], x, y);
  music();
  drawfull(final, x, y);
  add100(scotab[quality]);
} 
  

void moveem(int *timep)
{
  unsigned jst, leg, ndir;
  unsigned flags;
  coor coo;
  int i, monsc0, monsc2;
  mcb *mp, **acthead, *actmcb, **headcur, **headnoncleared, *p;
  int uncleared, lstval;
  ubyte (* const staga)[12] = playfield[player].staga;
  
  static unsigned around[16] = {0,2,1,3,8,0xa,9,0xb,4,6,5,7,0xc,0xe,0xd,0xf};
  *timep = -1;
  catchup();
  jst = d_joystc();
  if ((jst & 3) == 3)
    jst &= ~3;
  if ((jst & 0xc) == 0xc)
    jst &= ~0xc;
  if (demflg) {
    if (jst & 8)
      longjmp(gamebuf,QUIT);
    if (jst & 4) {
      demflg = 0;
      longjmp(gamebuf,GAME);
    }
    if (--demcnt)
      jst = 0;
    else {
      demcnt = dempnt[0];
      jst = dempnt[1] & 0xf;
      fired = (dempnt[1] & 0x80) != 0;
      if (dempnt[1] & 0x40)
	killall();
      dempnt += 2;
    }
  }
  leg = legal(&mymcb, staga);
  ndir = jst & ~mymcb.flag & leg;
  if (ndir) {
    if (ndir & (ndir-1)) {	/* more than one new legal direction */
      if (mymcb.flag & 0xc)
	ndir &= 3;
      else
	ndir &= 0xc;
    }
    mymcb.flag = (mymcb.flag & ~0xf) | ndir;
  } else
    mymcb.flag &= leg | ~0xf;
  for (mp = mymcb.next, i=nmons; i > 0; --i, mp=mp->next) {
    if (mp->flag & (1<<DEATH) || mp->downcn > 1)
      continue;
    leg = legal(mp, staga);
    if (leg & (leg-1)) {
      leg &= ~around[mp->flag & 0xf];
      if (leg & (leg-1)) {
	if (mp->flag & 3) {
	  if (mymcb.ycoor < mp->ycoor) {
	    if (leg & 4)
	      mp->flag = (mp->flag & ~0xf) | 4;
	    else if (!(leg & mp->flag)) {
	      assert(leg & 8);
	      mp->flag = (mp->flag & ~0xf) | 8;
	    }
	  } else if (leg & 8)
	    mp->flag = (mp->flag & ~0xf) | 8;
	  else if (!(leg & mp->flag)) {
	    assert(leg & 4);
	    mp->flag = (mp->flag & ~0xf) | 4;
	  }
	} else if (mymcb.xcoor < mp->xcoor) {
	  if (leg & 2)
	    mp->flag = (mp->flag & ~0xf) | 2;
	  else if (!(leg & mp->flag)) {
	    assert(leg & 1);
	    mp->flag = (mp->flag & ~0xf) | 1;
	  }
	} else if (leg & 1)
	  mp->flag = (mp->flag & ~0xf) | 1;
	else if (!(leg & mp->flag)) {
	  assert(leg & 2);
	  mp->flag = (mp->flag & ~0xf) | 2;
	}
      } else
	mp->flag = (mp->flag & ~0xf) | leg;
    } else 
      mp->flag = (mp->flag & ~0xf) | leg;  
  }
  if (mymcb.flag & 1) {
    if (mymcb.xcoor == scrx+26
	&& scrx < (playfield[player].stage.width-12)*4)
      scrlft();
  } else if (mymcb.flag & 2) {
    if (mymcb.xcoor == scrx+18
	&& scrx > 0)
      scrrig();
  }
  flags = newcoo(&coo, &mymcb);
  if (flags & B_MOVED) {
    if (!((coo.x | coo.y)&3)) {
      switch(staga[coo.x/4][coo.y/4]) {
      case AO:
	do {
	  coo.y += 4;
	} while (staga[coo.x/4][coo.y/4] != AU);
	break;
      case AU:
	do {
	  coo.y -= 4;
	} while (staga[coo.x/4][coo.y/4] != AO);
	break;
      }
    }
  }
  lstval = 0;
  monsc0 = nmons+1;		/* downcounter redraw */
  monsc2 = 0;			/* number of precleared monsters */
  headnoncleared = 0;		/* head of those after precleared */
  lstval = 0;			/*  */
  actmcb = &mymcb;		/* current monster */
  acthead = 0;			/* link to current monster */
  if (monsc0--) {
    for  (;;) {
      if ((uncleared = (monsc2 == 0)) != 0)
	headnoncleared = &actmcb->next; /* current monster not yet cleared */
      else
	--monsc2;
      headcur = headnoncleared;
      for (i=monsc0-monsc2, lstval=0; i--;) {
	p = *headcur;
	if (   (actmcb->xcoor+4 > p->xcoor
		&& actmcb->xcoor < p->xcoor+4
		&& actmcb->ycoor+4 > p->ycoor
		&& actmcb->ycoor < p->ycoor+4)
	       || (flags & B_MOVED
		   && coo.x+4 > p->xcoor
		   && coo.x < p->xcoor+4
		   && coo.y+4 > p->ycoor
		   && coo.y < p->ycoor+4)
	       ) {
	  grabit(p);
	  ++monsc2;
	  if (lstval) {
	    *headcur = p->next;
	    p->next = *headnoncleared;
	    *headnoncleared = p;
	    headnoncleared = &p->next;
	  } else {
	    headcur = headnoncleared = &p->next;
	  }
	} else {
	  headcur = &p->next;
	  lstval = 1;
	}
      }
      /* Now all shapes intersecting with actmcb are cleared */
      if (flags & B_MOVED) {
	if (coo.x + 6 < scrx
	    || coo.x > scrx + 50 )
	  flags |= B_KILLED;
	if (uncleared)
	  grabit(actmcb);
	if (flags & B_KILLED) {
	  *acthead = actmcb->next;
	  actmcb->next = freemcb;
	  freemcb = actmcb;
	  actmcb = *acthead;
	  --nmons;
	  if (!monsc0--)
	    break;
	  flags = newcoo(&coo, actmcb);
	  continue;
	} else {
	  if (flags & B_CLOUDY) {
	    actmcb->shape = d_cloudshapes;
	    if (actmcb->downcn > 2)
	      actmcb->flag &= ~(1<<PHASE);
	    else
	      actmcb->flag |= 1<<PHASE;
	  } else if (flags & B_BORN)
	    actmcb->shape = actmcb->mph->shapri;
	  else {
	    actmcb->flag ^= (1<<PHASE);
	    actmcb->xcoor = coo.x;
	    actmcb->ycoor = coo.y;
	  }
	  throwi(actmcb);
	}
      } else if (!uncleared)
	throwi(actmcb);
      if (!monsc0--)
	break;
      acthead = &actmcb->next;
      actmcb = *acthead;
      flags = newcoo(&coo, actmcb);
    }
  }
  if (((mymcb.xcoor | mymcb.ycoor)&3) == 0) {
    switch (staga[mymcb.xcoor/4][mymcb.ycoor/4]) {
    case BO:
      staga[mymcb.xcoor/4][mymcb.ycoor/4] = GA;
      staga[mymcb.xcoor/4][mymcb.ycoor/4-1] = OX;
      sndbon();
      gbonus(playfield[player].nbonus++, mymcb.xcoor, mymcb.ycoor-4, oxshd);
      break;
    case BU:
      staga[mymcb.xcoor/4][mymcb.ycoor/4] = GA;
      staga[mymcb.xcoor/4][mymcb.ycoor/4+1] = UX;
      sndbon();
      gbonus(playfield[player].nbonus++, mymcb.xcoor, mymcb.ycoor+4, uxshd);
      break;
    case SR:
      if (player[playfield].wkey)
	break;
      player[playfield].wkey=1;
      staga[mymcb.xcoor/4][mymcb.ycoor/4] = GA;
      staga[mymcb.xcoor/4+1][mymcb.ycoor/4] = SX;
      sndkey();
      gbonus(1, mymcb.xcoor+4, mymcb.ycoor, sxshd);
      mymcb.mph = &manwmph;
      throwi(&mymcb);
      break;
    case LR:
      if (!player[playfield].wkey)
	break;
      player[playfield].wkey = 0;
      staga[mymcb.xcoor/4][mymcb.ycoor/4] = GA;
      sndloc();
      music();
      for (i=0; i<4; i++) {
	d_liftgates(mymcb.xcoor+4-scrx, mymcb.ycoor, colbak);
	delay(14);
	music();
      }
      mymcb.mph = &manmph;
      grabit(&mymcb);
      throwi(&mymcb);
      add100(10);
      switch (staga[mymcb.xcoor/4+2][mymcb.ycoor/4]) {
      case T4:
	longjmp(gamebuf, WIN);
      case LO:
	staga[mymcb.xcoor/4+1][mymcb.ycoor/4] = LR;
	break;
      default:
	staga[mymcb.xcoor/4+1][mymcb.ycoor/4] = GA;
	break;
      }
      break;
    }
  }
  for (actmcb = mymcb.next, i=0; i<nmons; ++i, actmcb = actmcb->next) {
    if (!(actmcb->flag & ((1<<BIRTH)|(1<<DEATH)))
	&& mymcb.xcoor + 2 >= actmcb->xcoor
	&& mymcb.xcoor <= actmcb->xcoor + 2
	&& mymcb.ycoor + 2 >= actmcb->ycoor
	&& mymcb.ycoor <= actmcb->ycoor + 2) {
      drawfull(deadshd, mymcb.xcoor, mymcb.ycoor);
      snddea();
      music();
      d_timeclear();
      longjmp(gamebuf, LOSE);
    }
  }
}
